@@ -81,7 +81,7 @@ INIT {
    }
 }
 
-our $VERSION = '1.201_004';
+our $VERSION = '1.202_001';
 $VERSION = eval $VERSION;
 
 ###############################################################################
@@ -101,7 +101,7 @@ use constant {
    QUE_TEMPLATE     => $_que_template,   ## Pack template for queue socket
    QUE_READ_SIZE    => $_que_read_size,  ## Read size
 
-   OUTPUT_W_DNE     => ':W~DNE',         ## Worker has completed
+   OUTPUT_W_DNE     => ':W~DNE',         ## Worker has completed job
    OUTPUT_W_EXT     => ':W~EXT',         ## Worker has exited
    OUTPUT_A_ARY     => ':A~ARY',         ## Array  << Array
    OUTPUT_S_GLB     => ':S~GLB',         ## Scalar << Glob FH
@@ -133,22 +133,23 @@ undef $_max_files; undef $_max_procs;
 undef $_que_template; undef $_que_read_size;
 
 my %_valid_fields = map { $_ => 1 } qw(
-   max_workers tmp_dir use_threads user_tasks task_end
+   chunk_size input_data max_workers use_slurpio use_threads
+   flush_file flush_stderr flush_stdout stderr_file stdout_file
+   job_delay spawn_delay submit_delay tmp_dir user_tasks task_end
+   user_begin user_end user_func user_error user_output
+   on_exit_post on_run_post
 
-   chunk_size input_data job_delay spawn_delay submit_delay use_slurpio
-   flush_file flush_stderr flush_stdout stderr_file stdout_file on_post_exit
-   user_begin user_end user_func user_error user_output on_post_run
-
-   _abort_msg _mce_sid _mce_tid _pids _run_mode _single_dim _thrs _tids _wid
+   _mce_sid _mce_tid _pids _sess_dir _spawned _task0_max_workers _thrs _tids
    _com_r_sock _com_w_sock _dat_r_sock _dat_w_sock _out_r_sock _out_w_sock
-   _que_r_sock _que_w_sock _sess_dir _spawned _state _status _task _task_id
-   _exiting _exit_pid _total_exited _total_workers
+   _que_r_sock _que_w_sock _abort_msg _run_mode _single_dim _task_id _wid
+   _exiting _exit_pid _state _status _total_exited _total_workers
 );
 
 my %_params_allowed_args = map { $_ => 1 } qw(
    chunk_size input_data job_delay spawn_delay submit_delay use_slurpio
-   flush_file flush_stderr flush_stdout stderr_file stdout_file on_post_exit
-   user_begin user_end user_func user_error user_output on_post_run
+   flush_file flush_stderr flush_stdout stderr_file stdout_file
+   user_begin user_end user_func user_error user_output
+   on_exit_post on_run_post
 );
 
 my $_is_cygwin   = ($^O eq 'cygwin');
@@ -209,8 +210,8 @@ sub new {
    $self->{job_delay}    = $argv{job_delay}    || undef;
    $self->{spawn_delay}  = $argv{spawn_delay}  || undef;
    $self->{submit_delay} = $argv{submit_delay} || undef;
-   $self->{on_post_exit} = $argv{on_post_exit} || undef;
-   $self->{on_post_run}  = $argv{on_post_run}  || undef;
+   $self->{on_exit_post} = $argv{on_exit_post} || undef;
+   $self->{on_run_post}  = $argv{on_run_post}  || undef;
    $self->{user_begin}   = $argv{user_begin}   || undef;
    $self->{user_func}    = $argv{user_func}    || undef;
    $self->{user_end}     = $argv{user_end}     || undef;
@@ -259,7 +260,6 @@ sub new {
    $self->{_que_r_sock} = undef; ## Queue channel for MCE
    $self->{_que_w_sock} = undef; ## Queue channel for workers
    $self->{_state}      = undef; ## State info: worker (task_id/task/params)
-   $self->{_task}       = undef; ## Task info: max_workers
 
    ## -------------------------------------------------------------------------
 
@@ -362,22 +362,23 @@ sub spawn {
    socketpair( $self->{_com_r_sock}, $self->{_com_w_sock},
       AF_UNIX, SOCK_STREAM, PF_UNSPEC ) or die "socketpair: $!\n";
 
-   binmode  $self->{_com_r_sock};                 ## Set binary mode
-   binmode  $self->{_com_w_sock};
+   binmode $self->{_com_r_sock};                  ## Set binary mode
+   binmode $self->{_com_w_sock};
 
    ## Create socket pair for data channels between MCE and workers.
    socketpair( $self->{_dat_r_sock}, $self->{_dat_w_sock},
       AF_UNIX, SOCK_STREAM, PF_UNSPEC ) or die "socketpair: $!\n";
 
-   binmode  $self->{_dat_r_sock};                 ## Set binary mode
-   binmode  $self->{_dat_w_sock};
+   binmode $self->{_dat_r_sock};                  ## Set binary mode
+   binmode $self->{_dat_w_sock};
 
    ## Create socket pair for serializing send requests and STDOUT output.
    socketpair( $self->{_out_r_sock}, $self->{_out_w_sock},
       AF_UNIX, SOCK_STREAM, PF_UNSPEC ) or die "socketpair: $!\n";
 
-   binmode  $self->{_out_r_sock};                 ## Set binary mode
-   binmode  $self->{_out_w_sock};
+   binmode $self->{_out_r_sock};                  ## Set binary mode
+   binmode $self->{_out_w_sock};
+
    shutdown $self->{_out_r_sock}, 1;              ## No more writing
    shutdown $self->{_out_w_sock}, 0;              ## No more reading
 
@@ -385,8 +386,9 @@ sub spawn {
    socketpair( $self->{_que_r_sock}, $self->{_que_w_sock},
       AF_UNIX, SOCK_STREAM, PF_UNSPEC ) or die "socketpair: $!\n";
 
-   binmode  $self->{_que_r_sock};                 ## Set binary mode
-   binmode  $self->{_que_w_sock};
+   binmode $self->{_que_r_sock};                  ## Set binary mode
+   binmode $self->{_que_w_sock};
+
    shutdown $self->{_que_r_sock}, 1;              ## No more writing
    shutdown $self->{_que_w_sock}, 0;              ## No more reading
 
@@ -408,7 +410,7 @@ sub spawn {
    my $_wid = 0;
 
    $self->{_pids}   = (); $self->{_thrs}  = (); $self->{_tids} = ();
-   $self->{_status} = (); $self->{_state} = (); $self->{_task} = ();
+   $self->{_status} = (); $self->{_state} = ();
 
    ## Obtain lock.
    open my $_COM_LOCK, '+>> :stdio', "$_sess_dir/com.lock";
@@ -424,12 +426,14 @@ sub spawn {
          $self->_dispatch_child(++$_wid) for (1 .. $_max_workers);
       }
 
-      $self->{_task}->[0] = { _total_workers => $_max_workers };
+      $self->{_task0_max_workers} = $_max_workers;
       $_wid = 0;
 
-      for (1 .. $_max_workers) { $self->{_state}->[++$_wid] = {
-         _task_id => undef, _task => undef, _params => undef
-      }}
+      for (1 .. $_max_workers) {
+         $self->{_state}->[++$_wid] = {
+            _task_id => undef, _task => undef, _params => undef
+         }
+      }
    }
    else {
       if ($_is_cygwin) {
@@ -467,23 +471,21 @@ sub spawn {
          $_task_id++;
       }
 
+      $self->{_task0_max_workers} = $self->{user_tasks}->[0]->{max_workers};
       $_wid = $_task_id = 0;
 
       for my $_task (@{ $self->{user_tasks} }) {
-         $self->{_task}->[$_task_id] = {
-            _total_workers => $_task->{max_workers}
-         };
-         for (1 .. $_task->{max_workers}) { $self->{_state}->[++$_wid] = {
-            _task_id => $_task_id, _task => $_task, _params => undef
-         }}
-
+         for (1 .. $_task->{max_workers}) {
+            $self->{_state}->[++$_wid] = {
+               _task_id => $_task_id, _task => $_task, _params => undef
+            }
+         }
          $_task_id++;
       }
    }
 
    ## Release lock.
    flock $_COM_LOCK, LOCK_UN;
-   select(undef, undef, undef, 0.003);
    close $_COM_LOCK; undef $_COM_LOCK;
 
    $SIG{__DIE__}  = $_die_handler;
@@ -514,9 +516,12 @@ sub foreach {
    my ($_user_func, $_params_ref);
 
    if (ref $_[2] eq 'HASH') {
-      $_user_func = $_[3]; $_params_ref = $_[2];
-   } else {
-      $_user_func = $_[2]; $_params_ref = {};
+      $_user_func  = $_[3];
+      $_params_ref = $_[2];
+   }
+   else {
+      $_user_func  = $_[2];
+      $_params_ref = {};
    }
 
    @_ = ();
@@ -553,9 +558,12 @@ sub forchunk {
    my ($_user_func, $_params_ref);
 
    if (ref $_[2] eq 'HASH') {
-      $_user_func = $_[3]; $_params_ref = $_[2];
-   } else {
-      $_user_func = $_[2]; $_params_ref = {};
+      $_user_func  = $_[3];
+      $_params_ref = $_[2];
+   }
+   else {
+      $_user_func  = $_[2];
+      $_params_ref = {};
    }
 
    @_ = ();
@@ -593,7 +601,8 @@ sub process {
    ## Set input data.
    if (defined $_input_data) {
       $_params_ref->{input_data} = $_input_data;
-   } else {
+   }
+   else {
       _croak("MCE::process: 'input_data' is not specified")
    }
 
@@ -631,13 +640,12 @@ sub restart_worker {
    my $_use_threads = (defined $_task->{use_threads})
       ? $_task->{use_threads} : $self->{use_threads};
 
-   $self->{_task}->[$_task_id]->{_total_workers} += 1 if (defined $_task_id);
-   $self->{_total_exited} -= 1; $self->{_total_workers} += 1;
-
    if (defined $_use_threads && $_use_threads == 1) {
       $self->_dispatch_thread($_wid, $_task, $_task_id, $_params);
+      $self->{_total_exited} -= 1; $self->{_total_workers} += 1;
    } else {
       $self->_dispatch_child($_wid, $_task, $_task_id, $_params);
+      $self->{_total_exited} -= 1; $self->{_total_workers} += 1;
    }
 
    return;
@@ -772,28 +780,20 @@ sub run {
       '_max_workers' => $_max_workers
    );
 
-   my $_COM_LOCK;
-
    ## Begin processing.
    {
       local $\ = undef; local $/ = $LF;
       lock $_MCE_LOCK if ($_has_threads);            ## Obtain MCE lock.
 
-      my ($_wid, %_task0_wids);
+      my $_COM_R_SOCK    = $self->{_com_r_sock};
+      my $_submit_delay  = $self->{submit_delay};
+      my $_wid;
 
-      my $_COM_R_SOCK     = $self->{_com_r_sock};
-      my $_submit_delay   = $self->{submit_delay};
-      my $_has_user_tasks = (defined $self->{user_tasks});
-
-      my $_frozen_params  = freeze(\%_params);
-      my $_frozen_nodata  = freeze(\%_params_nodata)
+      my $_frozen_params = freeze(\%_params);
+      my $_frozen_nodata = freeze(\%_params_nodata)
          if (defined $self->{user_tasks});
 
-      if ($_has_user_tasks) { for (1 .. @{ $self->{_state} } - 1) {
-         $_task0_wids{$_} = 1 unless ($self->{_state}->[$_]->{_task_id});
-      }}
-
-      ## Obtain lock 1 of 2.
+      ## Obtain lock.
       open my $_DAT_LOCK, '+>> :stdio', "$_sess_dir/dat.lock";
       flock $_DAT_LOCK, LOCK_EX;
 
@@ -812,7 +812,7 @@ sub run {
          print $_COM_R_SOCK $_, $LF;
          chomp($_wid = <$_COM_R_SOCK>);
 
-         if (!$_has_user_tasks || exists $_task0_wids{$_wid}) {
+         if ($_wid <= $self->{_task0_max_workers}) {
             print $_COM_R_SOCK length($_frozen_params), $LF, $_frozen_params;
             $self->{_state}->[$_wid]->{_params} = \%_params;
          } else {
@@ -823,25 +823,20 @@ sub run {
          <$_COM_R_SOCK>;
       }
 
-      ## Obtain lock 2 of 2.
-      open $_COM_LOCK, '+>> :stdio', "$_sess_dir/com.lock";
-      flock $_COM_LOCK, LOCK_EX;
+      select(undef, undef, undef, $_submit_delay)
+         if ($_submit_delay && $_submit_delay > 0.0);
 
-      ## Release lock 1 of 2.
+      ## Release lock.
       flock $_DAT_LOCK, LOCK_UN;
-      select(undef, undef, undef, 0.003);
       close $_DAT_LOCK; undef $_DAT_LOCK;
    }
 
    ## -------------------------------------------------------------------------
 
-   $self->{_total_workers} = $_max_workers;
-   $self->{_total_exited}  = 0;
-
    ## Call the output function.
    if ($_max_workers > 0) {
       $self->{_abort_msg} = $_abort_msg;
-      $self->_output_loop($_input_data, $_input_glob);
+      $self->_output_loop($_max_workers, $_input_data, $_input_glob);
       undef $self->{_abort_msg};
    }
 
@@ -855,13 +850,8 @@ sub run {
       }
    }
 
-   ## Release lock 2 of 2.
-   flock $_COM_LOCK, LOCK_UN;
-   select(undef, undef, undef, 0.003);
-   close $_COM_LOCK; undef $_COM_LOCK;
-
-   ## Shutdown workers (also if one or more workers have exited).
-   $self->shutdown() if ($_auto_shutdown == 1 || $self->{_total_exited} > 0);
+   ## Shutdown workers.
+   $self->shutdown() if ($_auto_shutdown == 1);
 
    return $self;
 }
@@ -914,8 +904,6 @@ sub shutdown {
    }
 
    flock $_DAT_LOCK, LOCK_UN;
-   select(undef, undef, undef, 0.003);
-   close $_DAT_LOCK; undef $_DAT_LOCK;
 
    ## Reap children/threads.
    if ( $self->{_pids} && @{ $self->{_pids} } > 0 ) {
@@ -931,6 +919,8 @@ sub shutdown {
       }
    }
 
+   close $_DAT_LOCK; undef $_DAT_LOCK;
+
    ## Close sockets.
    shutdown $self->{_out_r_sock}, 0;              ## Output channels
    shutdown $self->{_out_w_sock}, 1;
@@ -956,14 +946,14 @@ sub shutdown {
    if (defined $_sess_dir) {
       unlink "$_sess_dir/com.lock";
       unlink "$_sess_dir/dat.lock";
-      rmdir  "$_sess_dir";
+      rmdir $_sess_dir;
    }
 
    ## Reset vars.
    $self->{_total_exited} = $self->{_total_workers} = 0;
 
    $self->{_pids}   = (); $self->{_thrs}  = (); $self->{_tids} = ();
-   $self->{_status} = (); $self->{_state} = (); $self->{_task} = ();
+   $self->{_status} = (); $self->{_state} = ();
 
    $self->{_out_r_sock} = $self->{_out_w_sock} = undef;
    $self->{_que_r_sock} = $self->{_que_w_sock} = undef;
@@ -1018,7 +1008,7 @@ sub exit {
    _croak("MCE::exit: method cannot be called by the manager process")
       unless ($self->wid());
 
-   delete $_mce_spawned{ $self->{_mce_sid} };
+   $SIG{__DIE__} = $SIG{__WARN__} = sub { };
 
    unless ($self->{_exiting}) {
       $self->{_exiting} = 1;
@@ -1028,32 +1018,27 @@ sub exit {
       my $_task_id    = $self->{_task_id};
 
       if (defined $_DAT_LOCK) {
-         my $_len  =  length($_exit_msg);
-         $_exit_id =~ s/[\r\n][\r\n]*/ /mg;
+         my $_len = length($_exit_msg);
 
          local $\ = undef;
          flock $_DAT_LOCK, LOCK_EX;
 
-         flock $_DAT_LOCK, LOCK_EX          if ($_is_cygwin);  ## Humm, needed.
-         select(undef, undef, undef, 0.001) if ($_is_cygwin);  ## This as well.
-
-         print $_OUT_W_SOCK OUTPUT_W_EXT,  $LF,
-            ((defined $_task_id) ? "$_task_id${LF}" : "-1${LF}");
-
+         print $_OUT_W_SOCK OUTPUT_W_EXT,  $LF;
          print $_DAT_W_SOCK $self->wid(),  $LF, $self->{_exit_pid}, $LF,
                             $_exit_status, $LF, $_exit_id,          $LF,
-                            $_len,         $LF, $_exit_msg;
+                            $_len,         $LF;
 
+         print $_DAT_W_SOCK $_exit_msg          if ($_len);
+
+         print $_OUT_W_SOCK OUTPUT_W_DNE,  $LF;
+         print $_DAT_W_SOCK $_task_id,     $LF  if (defined $_task_id);
+
+         delete $_mce_spawned{ $self->{_mce_sid} };
          flock $_DAT_LOCK, LOCK_UN;
       }
    }
 
    ## Exit thread/child process.
-   $SIG{__DIE__} = $SIG{__WARN__} = sub { };
-
-   close $_DAT_LOCK; close $_COM_LOCK;
-   undef $_DAT_LOCK; undef $_COM_LOCK;
-
    threads->exit($_exit_status) if ($_has_threads && threads->can('exit'));
 
    close STDERR; close STDOUT;
@@ -1189,6 +1174,10 @@ sub do {
 ##
 ###############################################################################
 
+sub _NOOP {
+
+}
+
 sub _croak {
 
    $SIG{__DIE__}  = \&_die;
@@ -1199,10 +1188,15 @@ sub _croak {
    return;
 }
 
-sub _NOOP { }
+sub _die {
 
-sub _die  { MCE::Signal->_die_handler(@_);  }
-sub _warn { MCE::Signal->_warn_handler(@_); }
+   MCE::Signal->_die_handler(@_);
+}
+
+sub _warn {
+
+   MCE::Signal->_warn_handler(@_);
+}
 
 ###############################################################################
 ## ----------------------------------------------------------------------------
@@ -1270,10 +1264,10 @@ sub _validate_args {
    _croak "$_tag: 'submit_delay' is not valid"
       if ($_s->{submit_delay} && $_s->{submit_delay} !~ /\A[\d\.]+\z/);
 
-   _croak "$_tag: 'on_post_exit' is not a CODE reference"
-      if ($_s->{on_post_exit} && ref $_s->{on_post_exit} ne 'CODE');
-   _croak "$_tag: 'on_post_run' is not a CODE reference"
-      if ($_s->{on_post_run} && ref $_s->{on_post_run} ne 'CODE');
+   _croak "$_tag: 'on_exit_post' is not a CODE reference"
+      if ($_s->{on_exit_post} && ref $_s->{on_exit_post} ne 'CODE');
+   _croak "$_tag: 'on_run_post' is not a CODE reference"
+      if ($_s->{on_run_post} && ref $_s->{on_run_post} ne 'CODE');
    _croak "$_tag: 'user_begin' is not a CODE reference"
       if ($_s->{user_begin} && ref $_s->{user_begin} ne 'CODE');
    _croak "$_tag: 'user_func' is not a CODE reference"
@@ -1523,21 +1517,21 @@ sub _validate_args {
    my ($_I_SEP, $_O_SEP, $_input_glob, $_chunk_size);
    my ($_input_size, $_offset_pos, $_single_dim, $_use_slurpio);
 
-   my ($_has_user_tasks, $_on_post_exit, $_on_post_run, $_task_id);
+   my ($_has_user_tasks, $_task_id, @_task_max_workers);
    my ($_exit_wid, $_exit_pid, $_exit_status, $_exit_id);
+   my ($_on_exit_post, $_on_run_post);
 
    ## Create hash structure containing various output functions.
    my %_output_function = (
 
-      OUTPUT_W_DNE.$LF => sub {                   ## Worker has completed
+      OUTPUT_W_DNE.$LF => sub {                   ## Worker has completed job
          $self->{_total_workers} -= 1;
-         chomp($_task_id = <$_OUT_R_SOCK>);
 
-         ## Call on task_end if the last worker for the task.
-         if ($_has_user_tasks && $_task_id >= 0) {
-            $self->{_task}->[$_task_id]->{_total_workers} -= 1;
+         if ($_has_user_tasks) {
+            chomp($_task_id = <$_DAT_R_SOCK>);
+            $_task_max_workers[$_task_id] -= 1;
 
-            unless ($self->{_task}->[$_task_id]->{_total_workers}) {
+            unless ($_task_max_workers[$_task_id]) {
                if (defined $self->{user_tasks}->[$_task_id]->{task_end}) {
                   $self->{user_tasks}->[$_task_id]->{task_end}();
                }
@@ -1550,14 +1544,7 @@ sub _validate_args {
       ## ----------------------------------------------------------------------
 
       OUTPUT_W_EXT.$LF => sub {                   ## Worker has exited
-         $self->{_total_exited}  += 1;
-         $self->{_total_workers} -= 1;
-
-         chomp($_task_id = <$_OUT_R_SOCK>);
-
-         $self->{_task}->[$_task_id]->{_total_workers} -= 1
-            if ($_has_user_tasks && $_task_id >= 0);
-
+         $self->{_total_exited} += 1;
          my $_exit_msg = '';
 
          chomp($_exit_wid    = <$_DAT_R_SOCK>);
@@ -1591,15 +1578,15 @@ sub _validate_args {
             }
          }
 
-         ## Call on_post_exit callback if defined. Otherwise, append status
-         ## information if on_post_run is defined for later retrieval.
-         if (defined $_on_post_exit) {
-            $_on_post_exit->(
+         ## Call on_exit_post callback if defined. Otherwise, append status
+         ## information if on_run_post is defined for later retrieval.
+         if (defined $_on_exit_post) {
+            $_on_exit_post->(
                $self, $_exit_wid, $_exit_pid, $_exit_status,
                $_exit_msg, $_exit_id
             );
          }
-         elsif (defined $_on_post_run) {
+         elsif (defined $_on_run_post) {
             push @{ $self->{_status} }, {
                wid    => $_exit_wid,
                pid    => $_exit_pid,
@@ -1609,15 +1596,6 @@ sub _validate_args {
             };
          }
 
-         ## Call on task_end if the last worker for the task.
-         if ($_has_user_tasks && $_task_id >= 0) {
-            unless ($self->{_task}->[$_task_id]->{_total_workers}) {
-               if (defined $self->{user_tasks}->[$_task_id]->{task_end}) {
-                  $self->{user_tasks}->[$_task_id]->{task_end}();
-               }
-            }
-         }
-
          return;
       },
 
@@ -1876,16 +1854,17 @@ sub _validate_args {
 
    sub _output_loop {
 
-      $self        = $_[0];
-      $_input_data = $_[1];
-      $_input_glob = $_[2];
+      $self           = $_[0];
+      $_total_workers = $_[1];
+      $_input_data    = $_[2];
+      $_input_glob    = $_[3];
 
       @_ = ();
 
       die "Private method called" unless (caller)[0]->isa( ref($self) );
 
-      $_on_post_exit = $self->{on_post_exit};
-      $_on_post_run  = $self->{on_post_run};
+      $_on_exit_post = $self->{on_exit_post};
+      $_on_run_post  = $self->{on_run_post};
       $_chunk_size   = $self->{chunk_size};
       $_flush_file   = $self->{flush_file};
       $_max_workers  = $self->{max_workers};
@@ -1895,7 +1874,15 @@ sub _validate_args {
       $_single_dim   = $self->{_single_dim};
 
       $_has_user_tasks = (defined $self->{user_tasks});
-      $_eof_flag = 0;
+
+      $self->{_total_workers} = $_total_workers;
+      $self->{_total_exited} = $_eof_flag = 0;
+
+      if ($_has_user_tasks) {
+         push @_task_max_workers, $_->{max_workers} for (
+            @{ $self->{user_tasks} }
+         );
+      }
 
       if (defined $_input_data && ref $_input_data eq 'ARRAY') {
          $_input_size = @$_input_data;
@@ -1961,8 +1948,10 @@ sub _validate_args {
          }
       }
 
-      ## Call on_post_run callback.
-      $_on_post_run->($self, $self->{_status}) if (defined $_on_post_run);
+      ## Call on_run_post callback.
+      $_on_run_post->($self, $self->{_status}) if (defined $_on_run_post);
+
+      @_task_max_workers = undef if ($_has_user_tasks);
 
       ## Close opened sendto file handles.
       for (keys %_sendto_fhs) {
@@ -1977,6 +1966,9 @@ sub _validate_args {
       close $_MCE_STDOUT; undef $_MCE_STDOUT;
       close $_MCE_STDERR; undef $_MCE_STDERR;
 
+      ## Shutdown remaining workers if one or more have exited.
+      $self->shutdown() if ($self->{_total_exited} > 0);
+
       return;
    }
 }
@@ -2311,9 +2303,12 @@ sub _worker_do {
 
    ## Notify the main process a worker has completed.
    local $\ = undef;
+   flock $_DAT_LOCK, LOCK_EX;
 
-   print $_OUT_W_SOCK OUTPUT_W_DNE, $LF,
-      (defined $_task_id) ? "$_task_id${LF}" : "-1${LF}";
+   print $_OUT_W_SOCK OUTPUT_W_DNE, $LF;
+   print $_DAT_W_SOCK $_task_id,    $LF  if (defined $_task_id);
+
+   flock $_DAT_LOCK, LOCK_UN;
 
    return;
 }
@@ -2370,9 +2365,11 @@ sub _worker_loop {
          $_params_ref = thaw($_buffer);
       }
 
-      ## Wait until MCE completes params submission to all workers.
-      flock $_DAT_LOCK, LOCK_SH;
-      flock $_DAT_LOCK, LOCK_UN;
+      ## Wait until MCE completes job submission.
+      if (defined $self->{user_tasks}) {
+         flock $_DAT_LOCK, LOCK_SH;
+         flock $_DAT_LOCK, LOCK_UN;
+      }
 
       ## Update ID. Process request.
       $self->{_wid} = $_response unless (defined $self->{user_tasks});
@@ -2382,10 +2379,6 @@ sub _worker_loop {
 
       $self->_worker_do($_params_ref);
       undef $_params_ref;
-
-      ## Wait until all workers completed processing.
-      flock $_COM_LOCK, LOCK_SH;
-      flock $_COM_LOCK, LOCK_UN;
    }
 
    ## Notify the main process a worker has ended. The following is executed
@@ -2424,11 +2417,6 @@ sub _worker_main {
       $self->exit(255, $_[0]);
    };
 
-   ## Use code references from user task if defined.
-   $self->{user_begin} = $_task->{user_begin} if (defined $_task->{user_begin});
-   $self->{user_func}  = $_task->{user_func}  if (defined $_task->{user_func});
-   $self->{user_end}   = $_task->{user_end}   if (defined $_task->{user_end});
-
    ## Init runtime vars. Obtain handle to lock files.
    my $_mce_sid  = $self->{_mce_sid};
    my $_sess_dir = $self->{_sess_dir};
@@ -2441,6 +2429,11 @@ sub _worker_main {
    open $_COM_LOCK, '+>> :stdio', "$_sess_dir/com.lock";
    open $_DAT_LOCK, '+>> :stdio', "$_sess_dir/dat.lock";
 
+   ## Use code references from user task if defined.
+   $self->{user_begin} = $_task->{user_begin} if (defined $_task->{user_begin});
+   $self->{user_func}  = $_task->{user_func}  if (defined $_task->{user_func});
+   $self->{user_end}   = $_task->{user_end}   if (defined $_task->{user_end});
+
    ## Define status ID.
    my $_use_threads = (defined $_task->{use_threads})
       ? $_task->{use_threads} : $self->{use_threads};
@@ -2454,29 +2447,31 @@ sub _worker_main {
    ## Undef vars not required after being spawned.
    $self->{_com_r_sock}  = $self->{_dat_r_sock}  = $self->{_out_r_sock} = undef;
    $self->{flush_stderr} = $self->{flush_stdout} = undef;
-   $self->{on_post_exit} = $self->{on_post_run}  = undef;
+   $self->{on_exit_post} = $self->{on_run_post}  = undef;
    $self->{stderr_file}  = $self->{stdout_file}  = undef;
    $self->{user_error}   = $self->{user_output}  = undef;
    $self->{flush_file}   = undef;
 
    $self->{_pids}   = (); $self->{_thrs}  = (); $self->{_tids} = ();
-   $self->{_status} = (); $self->{_state} = (); $self->{_task} = ();
+   $self->{_status} = (); $self->{_state} = ();
 
    foreach (keys %_mce_spawned) {
       delete $_mce_spawned{$_} unless ($_ eq $_mce_sid);
    }
 
-   ## Begin processing if a new worker was added while running.
-   if (defined $_params) {
+   ## Wait until MCE completes spawning.
+   unless (defined $_params) {
+      flock $_COM_LOCK, LOCK_SH;
+      flock $_COM_LOCK, LOCK_UN;
+   }
+
+   ## Otherwise, a new worker was spawned taking place of a previous worker.
+   else {
       $self->{_wid} = $_wid;
       $self->_worker_do($_params);
       undef $_params;
    }
 
-   ## Wait until MCE completes spawning or worker completes running.
-   flock $_COM_LOCK, LOCK_SH;
-   flock $_COM_LOCK, LOCK_UN;
-
    ## Enter worker loop.
    my $_status = $self->_worker_loop();
    delete $_mce_spawned{ $self->{_mce_sid} };
@@ -2485,12 +2480,12 @@ sub _worker_main {
    $SIG{__DIE__} = $SIG{__WARN__} = sub { };
 
    eval {
-      flock $_DAT_LOCK, LOCK_SH; flock $_DAT_LOCK, LOCK_UN;
+      flock $_DAT_LOCK, LOCK_SH;
+      flock $_DAT_LOCK, LOCK_UN;
+      close $_DAT_LOCK; undef $_DAT_LOCK;
+      close $_COM_LOCK; undef $_COM_LOCK;
    };
 
-   close $_DAT_LOCK; close $_COM_LOCK;
-   undef $_DAT_LOCK; undef $_COM_LOCK;
-
    return;
 }
 
@@ -2529,16 +2524,7 @@ sub _dispatch_child {
       CORE::exit();
    }
 
-   if (defined $_pid) {
-      ## Store into an available slot, otherwise append to array.
-      if (defined $_params) { for (0 .. @{ $self->{_pids} } - 1) {
-         unless (defined $self->{_pids}->[$_]) {
-            $self->{_pids}->[$_] = $_pid;
-            return;
-         }
-      }}
-      push @{ $self->{_pids} }, $_pid;
-   }
+   push @{ $self->{_pids} }, $_pid if (defined $_pid);
 
    return;
 }
@@ -2572,14 +2558,6 @@ sub _dispatch_thread {
       unless (defined $_thr);
 
    if (defined $_thr) {
-      ## Store into an available slot, otherwise append to arrays.
-      if (defined $_params) { for (0 .. @{ $self->{_tids} } - 1) {
-         unless (defined $self->{_tids}->[$_]) {
-            $self->{_thrs}->[$_] = \$_thr;
-            $self->{_tids}->[$_] = $_thr->tid();
-            return;
-         }
-      }}
       push @{ $self->{_thrs} }, \$_thr;
       push @{ $self->{_tids} }, $_thr->tid();
    }
